/*
common.c

Copyright (C) 1999 Lars Brinkhoff.  See COPYING for terms and conditions.

Code common to both htc and hts.
*/

#include <stdio.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdlib.h>
#include <termios.h>
#include <sys/poll.h>

#include "tunnel.h"
#include "common.h"

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#ifndef HAVE_VSYSLOG
void
vsyslog (int level, char *fmt0, ...)
{
  char buffer[512]; /* FIXME: buffer overrun */
  va_list ap;

  va_start (ap, fmt0);
  if (vsprintf (buffer, fmt0, ap) > 0)
    syslog (level, buffer);
  va_end (ap);
}
#endif /* HAVE_VSYSLOG */

#ifdef DEBUG
void
log_level (int level, char *fmt0, va_list ap)
{
  if (debug_level >= level)
    {
      int i;
      for (i = 0; i < level; i++)
	fprintf (debug_file, "    ");
      vfprintf (debug_file, fmt0, ap);
      fprintf (debug_file, "\n");
      fflush (debug_file);
    }
}
#endif

void
log_notice (char *fmt0, ...)
{
  va_list ap;
  va_start(ap, fmt0);
#ifdef DEBUG
  log_level (0, fmt0, ap);
#else
  vsyslog (LOG_NOTICE, fmt0, ap);
#endif
  va_end(ap);
}

void
log_error (char *fmt0, ...)
{
  va_list ap;
  va_start(ap, fmt0);
#ifdef DEBUG
  log_level (1, fmt0, ap);
#else
  vsyslog (LOG_ERROR, fmt0, ap);
#endif
  va_end(ap);
}

#ifdef DEBUG
void
log_debug (char *fmt0, ...)
{
  va_list ap;
  va_start(ap, fmt0);
  log_level (2, fmt0, ap);
  va_end(ap);
}
#endif

#ifdef DEBUG
void
log_verbose (char *fmt0, ...)
{
  va_list ap;
  va_start(ap, fmt0);
  log_level (3, fmt0, ap);
  va_end(ap);
}
#endif

#ifdef DEBUG
void
log_annoying (char *fmt0, ...)
{
  va_list ap;
  va_start (ap, fmt0);
  log_level (4, fmt0, ap);
  va_end(ap);
}
#endif

int
server_socket (int port)
{
  struct sockaddr_in address;
  int s;

  s = socket (PF_INET, SOCK_STREAM, 0);
  if (s == -1)
    return -1;
  
  address.sin_family = PF_INET;
  address.sin_port = htons (port);
  address.sin_addr.s_addr = INADDR_ANY;
  
  if (bind (s, (struct sockaddr *)&address, sizeof (address)) == -1)
    {
      close (s);
      return -1;
    }

  if (listen (s, 1) == -1)
    {
      close (s);
      return -1;
    } 

  return s;
}

int
set_address (struct sockaddr_in *address, const char *host, int port)
{
  address->sin_family = PF_INET;
  address->sin_port = htons (port);
  address->sin_addr.s_addr = inet_addr (host);

  if (address->sin_addr.s_addr == INADDR_NONE)
    {
      struct hostent *hostent;

      log_annoying ("set_address: gethostbyname (\"%s\")", host);
      hostent = gethostbyname (host);
      if (hostent == 0)
	return -1;

      memcpy(&address->sin_addr.s_addr,
	     hostent->h_addr,
	     hostent->h_length);
      log_annoying ("set_address: host = %d.%d.%d.%d",
		     address->sin_addr.s_addr >> 24,
		    (address->sin_addr.s_addr >> 16) & 0xff,
		    (address->sin_addr.s_addr >>  8) & 0xff,
		     address->sin_addr.s_addr        & 0xff);
    }

  return 0;
}

int
open_device (char *device)
{
  struct termios t;
  int fd;

  fd = open (device, O_RDWR | O_NONBLOCK);
  if (fd == -1)
    return -1;
  
  if (tcgetattr (fd, &t) == -1)
    return -1;
  t.c_iflag = 0;
  t.c_oflag = 0;
  t.c_lflag = 0;
  if (tcsetattr (fd, TCSANOW, &t) == -1)
    return -1;

  return fd;
}

#ifdef DEBUG
void
dump_buf (FILE *f, unsigned char *buf, size_t len)
{
  const int N = 20;
  int i, j;

  for (i = 0; i < len;)
    {
      fputc ('[', f);
      for (j = 0; j < N && i + j < len; j++)
	fprintf (f, "%02x", buf[i + j]);
      for (; j < N; j++)
	fprintf (f, "  ");
      fputc (']', f);
      fputc ('[', f);
      for (j = 0; j < N && i + j < len; j++)
	{
	  int c = buf[i + j];
	  if (c < ' ' || c > 126)
	    fputc ('.', f);
	  else
	    fputc (c, f);
	}
      fputc (']', f);
      fputc ('\n', f);
      i += j;
    }
}
#endif

int
handle_device_input (Tunnel *tunnel, int fd, int events)
{
  unsigned char buf[10240];
  ssize_t n, m;

  if (events & POLLIN)
    {
      n = read (fd, buf, sizeof buf);
      if (n == 0 || n == -1)
	{
	  if (n == -1 && errno != EAGAIN)
	    log_error ("handle_device_input: read() error: %s",
		       strerror (errno));
	  return n;
	}

#ifdef DEBUG
      log_annoying ("read %d bytes from device:", n);
      if (debug_level >= 4)
	dump_buf (debug_file, buf, n);
#endif

      m = tunnel_write (tunnel, buf, n);
      log_annoying ("tunnel_write (%p, %p, %d) = %d", tunnel, buf, n, m);
      return m;
    }
  else if (events & POLLHUP)
    log_error ("handle_device_input: PULLHUP");
  else if (events & POLLERR)
    log_error ("handle_device_input: PULLERR");
  else if (events & POLLNVAL)
    log_error ("handle_device_input: PULLINVAL");
  else
    log_error ("handle_device_input: none of the above");

  return -1;
}

int
handle_tunnel_input (Tunnel *tunnel, int fd, int events)
{
  unsigned char buf[10240];
  ssize_t n, m;

  if (events & POLLIN)
    {
      n = tunnel_read (tunnel, buf, sizeof buf);
      if (n == 0 || n == -1)
	{
	  if (n == -1 && errno != EAGAIN)
	    log_error ("handle_tunnel_input: tunnel_read() error: %s",
		       strerror (errno));
	  return n;
	}

#ifdef DEBUG
      log_annoying ("read %d bytes from tunnel:", n);
      if (debug_level >= 4)
	dump_buf (debug_file, buf, n);
#endif

      m = write_all (fd, buf, n);
      log_annoying ("write_all (%d, %p, %d) = %d", fd, buf, n, m);
      return m;
    }
  else if (events & POLLHUP)
    log_error ("handle_device_input: PULLHUP");
  else if (events & POLLERR)
    log_error ("handle_device_input: PULLERR");
  else if (events & POLLNVAL)
    log_error ("handle_device_input: PULLINVAL");
  else
    log_error ("handle_device_input: none of the above");

  return -1;
}

void
name_and_port (const char *nameport, char **name, int *port)
{
  char *p;

  *name = strdup (nameport);
  if (*name == NULL)
    {
      fprintf (stderr, "Out of memory\n");
      exit (1);
    }

  p = strchr (*name, ':');
  if (p != NULL)
    {
      *port = atoi (p + 1);
      *p = '\0';
    }
}

#ifndef HAVE_DAEMON
int
daemon (int nochdir, int noclose)
{
  if (fork () != 0)
    exit (0);

  if (!nochdir)
    chdir ("/");

  close (0);
  close (1);
  close (2);

  if (noclose == 0)
    {
      open ("/dev/null", O_RDONLY);
      open ("/dev/null", O_WRONLY);
      open ("/dev/null", O_WRONLY);
    }

  /* FIXME: disassociate from controlling terminal. */

  return 0;
}
#endif
