/*
http.c
Copyright (C) 1999 Lars Brinkhoff.  See COPYING for terms and conditions.
*/

#include <stdio.h>
#include <stdlib.h>

#include "http.h"
#include "common.h"

static inline int
http_method (int fd, Http_destination *dest,
	     const char *method, ssize_t length)
{
  char str[1000]; /* FIXME: possible buffer overflow */

  if (fd == -1)
    {
      log_error ("http_method: fd == -1");
      return -1;
    }

  if (dest->proxy_name == NULL)
    {
      if (length >= 0)
	{
	  sprintf (str,
		   "%s /index.html HTTP/1.1\r\n"
		   "Host: %s:%d\r\n"
		   "Content-Length: %d\r\n"
		   "Connection: close\r\n"
		   "\r\n",
		   method,
		   dest->host_name, dest->host_port,
		   length);
	}
      else
	{
	  sprintf (str,
		   "%s /index.html HTTP/1.1\r\n"
		   "Host: %s:%d\r\n"
		   "Connection: close\r\n"
		   "\r\n",
		   method,
		   dest->host_name, dest->host_port);
	}
    }
  else
    {
      if (length >= 0)
	{
	  sprintf (str,
		   "%s http://%s:%d/index.html HTTP/1.1\r\n"
		   "Host: %s:%d\r\n"
		   "Content-Length: %d\r\n"
		   "Connection: close\r\n"
		   "\r\n",
		   method,
		   dest->host_name, dest->host_port,
		   dest->host_name, dest->host_port,
		   length);
	}
      else
	{
	  sprintf (str,
		   "%s http://%s:%d/index.html HTTP/1.1\r\n"
		   "Host: %s:%d\r\n"
		   "Connection: close\r\n"
		   "\r\n",
		   method,
		   dest->host_name, dest->host_port,
		   dest->host_name, dest->host_port);
	}
    }

#ifdef DEBUG
  {
    char *s = strdup (str); /* FIXME: check for NULL */
    int n = strlen (s);
    int i;
      
    for (i = 0; i < n; i++)
      {
	if (s[i] == '\r')
	  s[i] = ';';
	else if (s[i] == '\n')
	  s[i] = ' ';
      }
    s[n - 2] = 0;
    log_verbose ("http_method: %s", s);
    free (s);
  }
#endif

  if (write_all (fd, str, strlen (str)) == -1)
    {
      log_error ("http_method: error writing header: %s");
      return -1;
    }

  return 0;
}

int
http_get (int fd, Http_destination *dest)
{
  return http_method (fd, dest, "GET", -1);
}

int
http_put (int fd, Http_destination *dest, size_t length)
{
  return http_method (fd, dest, "PUT", (ssize_t)length);
}

int
http_post (int fd, Http_destination *dest, size_t length)
{
  return http_method (fd, dest, "POST", (ssize_t)length);
}

int
http_error_to_errno (int err)
{
  /* Error codes taken from RFC2068. */
  switch (err)
    {
    case -1: /* system error */
      return errno;
    case -200: /* OK */
    case -201: /* Created */
    case -202: /* Accepted */
    case -203: /* Non-Authoritative Information */
    case -204: /* No Content */
    case -205: /* Reset Content */
    case -206: /* Partial Content */
      return 0;
    case -400: /* Bad Request */
      log_error ("http_error_to_errno: 400 bad request");
      return EIO;
    case -401: /* Unauthorized */
      log_error ("http_error_to_errno: 401 unauthorized");
      return EACCES;
    case -403: /* Forbidden */
      log_error ("http_error_to_errno: 403 forbidden");
      return EACCES;
    case -404: /* Not Found */
      log_error ("http_error_to_errno: 404 not found");
      return ENOENT;
    case -411: /* Length Required */
      log_error ("http_error_to_errno: 411 length required");
      return EIO;
    case -413: /* Request Entity Too Large */
      log_error ("http_error_to_errno: 413 request entity too large");
      return EIO;
    case -505: /* HTTP Version Not Supported       */
      log_error ("http_error_to_errno: 413 HTTP version not supported");
      return EIO;
    case -100: /* Continue */
    case -101: /* Switching Protocols */
    case -300: /* Multiple Choices */
    case -301: /* Moved Permanently */
    case -302: /* Moved Temporarily */
    case -303: /* See Other */
    case -304: /* Not Modified */
    case -305: /* Use Proxy */ 
    case -402: /* Payment Required */
    case -405: /* Method Not Allowed */
    case -406: /* Not Acceptable */
    case -407: /* Proxy Autentication Required */
    case -408: /* Request Timeout */
    case -409: /* Conflict */
    case -410: /* Gone */
    case -412: /* Precondition Failed */
    case -414: /* Request-URI Too Long */
    case -415: /* Unsupported Media Type */
    case -500: /* Internal Server Error */
    case -501: /* Not Implemented */
    case -502: /* Bad Gateway */
    case -503: /* Service Unavailable */
    case -504: /* Gateway Timeout */
      log_error ("http_error_to_errno: HTTP error %d", err);
      return EIO;
    default:
      log_error ("http_error_to_errno: unknown error %d", err);
      return EIO;
    }
}

static ssize_t
read_until (int fd, int ch, unsigned char *buf)
{
  ssize_t n, len;

  len = 0;
  while ((n = read_all (fd, buf + len, 1)) == 1)
    {
      if (buf[len] == ch)
	break;
      len++;
    }
  if (n <= 0)
    {
      log_error ("read_until: read error: %s", strerror (errno));
      return n;
    }

  return len;
}

int
http_parse_response (int fd, Http_response *response)
{
  unsigned char buf[100]; /* FIXME: buffer overflow */
  ssize_t n;
  int i;

  n = read_until (fd, '/', buf);
  if (n <= 0)
    return n;
  else if (n != 4 || memcmp (buf, "HTTP", 4) != 0)
    {
      log_error ("http_parse_response: expected \"HTTP\"");
      return -1;
    }

  n = read_until (fd, '.', buf);
  if (n <= 0)
    return n;
  else
    {
      buf[n] = 0;
      response->major_version = atoi (buf);
      log_verbose ("http_parse_response: major version = %d",
		   response->major_version);
    }

  n = read_until (fd, ' ', buf);
  if (n <= 0)
    return n;
  else
    {
      buf[n] = 0;
      response->minor_version = atoi (buf);
      log_verbose ("http_parse_response: minor version = %d",
		   response->minor_version);
    }

  n = read_until (fd, ' ', buf);
  if (n <= 0)
    return n;
  else
    {
      buf[n] = 0;
      response->status_code = atoi (buf);
      log_verbose ("http_parse_response: status code = %d",
		   response->status_code);
    }

  i = 0;
  while (i < 4)
    {
      const char *end_of_header = "\r\n\r\n";
      char c;

      n = read_all (fd, &c, 1);
      if (n != 1 && errno != EAGAIN)
        {
          log_error ("http_parse_response: error reading header: %s",
                     strerror (errno));
          return n;
        }

      if (c == end_of_header[i])
	i++;
      else
	i = 0;
    }

  return 1;
}
